/**
  *  \file game/sim.h
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SIM_H
#define PCC_V2_GAME_SIM_H

#include <vector>
#include "cpluslib/object.h"
#include "cpluslib/sigslot.h"
#include "game/spec.h"
#include "game/vcr.h"
#include "game/vcrdb.h"
#include "io/streams.h"

/*! \defgroup Simulator Combat Simulator
    Classes to manage data for the combat simulator, run the simulations,
    and evaluate them. */
/*@{*/

class GSimShip;
class GSimPlanet;
class GSimState;

class GSimObject : public Object {
    // common
    int  id;
    char name[20];
    char fcode[3];
    int  damage;
    int  shield;
    int  owner;
    int  experience_level;
    int  aggressiveness;
    int  flags;
    int32 flak_r_override;
    int  flak_c_override;

    bool changed;

    VcrStatItem stats;

 protected:
    void setChanged();

    friend class GSimState;
    void setUnchanged();
    bool isChanged() const;

 public:
    GSimObject();
    ~GSimObject();

    int  getId() const;
    void setId(int id);

    string_t getName() const;
    void     setName(string_t name);

    string_t getFCode() const;
    void     setFCode(string_t fcode);

    int  getDamage() const;
    void setDamage(int damage);

    int  getShield() const;
    void setShield(int shield);

    int  getOwner() const;
    void setOwner(int owner);

    int  getExperienceLevel() const;
    void setExperienceLevel(int experience_level);

    int  getFlags() const;
    void setFlags(int flags);

    int32 getFlakRatingOverride() const;
    void  setFlakRatingOverride(int32 r);

    int32 getFlakCompensationOverride() const;
    void  setFlakCompensationOverride(int32 r);

    VcrStatItem& getStat()
        { return stats; }
    const VcrStatItem& getStat() const
        { return stats; }

    virtual GSimShip*   asShip();
    virtual GSimPlanet* asPlanet();

    // derived information
    int getMaxDamage() const;
    int getMaxShield() const;

    /** Flag bits. */
    enum {
        fl_RandomFC       = 1,    ///< Friendly code randomisation enabled.
        fl_RandomFC1      = 2,    ///< Randomize first place of fcode.
        fl_RandomFC2      = 4,    ///< Randomize second place of fcode.
        fl_RandomFC3      = 8,    ///< Randomize third place of fcode.
        fl_RatingOverride = 16,   ///< Use FLAK rating overrides.
        fl_Cloaked        = 32,   ///< Ship is cloaked.
        fl_Deactivated    = 64    ///< Ignore this ship during simulation.
    };
};

class GSimShip : public GSimObject {
    // ships
    int  crew;
    int  hull;
    int  mass;                  /* custom ships */
    int  beam_type, beam_count;
    int  torp_type, torp_lcount;
    int  bay_count;
    int  ammo;
    int  engine_type;
    int  aggressiveness;
    int  intercept_id;

 public:
    GSimShip();
    GSimShip(const TSimShipData& file_data, int version);
    ~GSimShip();

    void storeData(TSimShipData& file_data, int version);

    int  getCrew() const;
    void setCrew(int crew);

    int  getHull() const;
    void setHull(int hull);

    int  getMass() const;
    void setMass(int mass);

    int  getBeamType() const;
    void setBeamType(int beam_type);
    int  getBeamCount() const;
    void setBeamCount(int beam_count);

    int  getTorpType() const;
    void setTorpType(int torp_type);
    int  getTorpLauncherCount() const;
    void setTorpLauncherCount(int torp_lcount);

    int  getBayCount() const;
    void setBayCount(int bay_count);

    int  getAmmo() const;
    void setAmmo(int ammo);

    int  getEngineType() const;
    void setEngineType(int engine_type);
    
    int  getAggressiveness() const;
    void setAggressiveness(int aggressiveness);

    int  getInterceptId() const;
    void setInterceptId(int id);

    GSimShip* asShip();

    // derived
    bool isCustomShip() const;

    bool isMatchingShipList() const;

    /** Aggressiveness values. Values between 1 and 12 correspond to primary
        enemy settings. */
    enum {
        /// "Kill" mission.
        agg_Kill    = -1,
        /// Passive unit.
        agg_Passive = 0,
        /// Unit has no fuel.
        agg_NoFuel  = 13
    };

};

class GSimPlanet : public GSimObject {
    // planets
    int  defense;
    int  base_defense;
    int  beam_tech;             /* 0 = no base */
    int  torp_tech;
    int  base_fighters;
    int  base_torps[NUM_TORPS];

 public:
    GSimPlanet();
    GSimPlanet(const TSimPlanetData& file_data, int version);
    ~GSimPlanet();

    void storeData(TSimPlanetData& file_data, int version);

    int  getDefense() const;
    void setDefense(int defense);

    int  getBaseDefense() const;
    void setBaseDefense(int base_defense);

    int  getBaseBeamTech() const;
    void setBaseBeamTech(int beam_tech);

    int  getBaseTorpTech() const;
    void setBaseTorpTech(int torp_tech);

    int  getBaseFighters() const;
    void setBaseFighters(int base_fighters);

    int  getBaseTorps(int tech) const;
    void setBaseTorps(int tech, int amount);

    int  getBaseTorpsAsType(int tech) const;

    bool hasBase() const
        { return getBaseBeamTech() != 0; }

    // FIXME!!!!!!!!!!!!!!!!!1111
    int  getBaseDamage() const
        { return 0; }
    void setBaseDamage(int)
        { }

    GSimPlanet* asPlanet();
};

struct GSimOptions {
    /** Balancing mode. Various algorithms to correct the TVCR
        left/right imbalance. */
    enum BalancingMode {
        BalanceNone,            ///< No balancing.
        Balance360k,            ///< 360kt bonus (HOST).
        BalanceMasterAtArms     ///< Master at Arms proposal.
    };
    /** VCR mode. Which VCR is used to run the simulation. */
    enum VcrMode {
        VcrHost,                ///< HOST.
        VcrPHost2,              ///< PHost 2.x.
        VcrPHost3,              ///< PHost 3.x.
        VcrPHost4,              ///< PHost 4.x.
        VcrFLAK                 ///< FLAK.
    };

    enum {
        BalanceMAX = BalanceMasterAtArms,
        VcrMAX = VcrFLAK
    };

    GPlayerBitMatrix alliance_settings;
    GPlayerBitMatrix enemy_settings;

    int            engine_shield_bonus;
    bool           scotty_bonus;
    bool           random_left_right;
    bool           honor_alliances;
    bool           only_one_simulation;
    bool           seed_control;
    BalancingMode  balancing_mode;
    VcrMode        vcr_mode;

    int            seed;

    GSimOptions();

    void loadDefaults(VcrMode mode);
};

class GSimState {
    std::vector<GSimShip> ships;
    GSimPlanet            planet;
    bool                  has_planet;
    bool                  structure_change;
 public:
    typedef std::vector<GSimShip>::size_type ship_index_t;
    
    GSimState();
    GSimState(const GSimState& other);
    ~GSimState();

    GSimState& operator=(const GSimState& other);

    bool hasPlanet() const
        { return has_planet; }
    void addPlanet();
    void removePlanet();

    ship_index_t addShip();
    ship_index_t getNumShips() const;
    void duplicateShip(ship_index_t ix, int id);
    void removeShip(ship_index_t ix);
    bool isShipIdInUse(int sid) const;
    const GSimShip* getShipById(int id) const;
    int getFreeId();

    bool isMatchingShipList() const;

    void initializeStats();

    void doUpdates();

    void load(Stream& in);
    void save(Stream& out);

    void swap(GSimState& other);

    GSimPlanet& getPlanet();
    const GSimPlanet& getPlanet() const;

    GSimShip& getShip(ship_index_t sid);
    const GSimShip& getShip(ship_index_t sid) const;

    Signal1<int> sig_ship_changed;
    Signal0      sig_planet_changed;
    Signal0      sig_structure_changed;

    enum {
        MAX_SHIPS = ::NUM_SHIPS
    };
};

/*@}*/

#endif
