/**
  *  \file gfx/gfxutil.h
  *  \brief Graphics Utilities
  *
  *  Utility classes used for implementing the graphics and UI system.
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFXUTIL_H
#define PCC_V2_GFX_GFXUTIL_H

#include <iosfwd>
#include <list>
#include <SDL_video.h>
#include "cpluslib/types.h"
#include "gfx/gfxtypes.h"

/** Fill Pattern. An 8x8 pattern usable to fill a rectangle.
    Consists of 8 bytes (one per line, first byte is topmost),
    the most-significant bit of each byte is leftmost. */
class GfxFillPattern {
    uint8 pat[8];  ///< Pattern data.
 public:
    GfxFillPattern();
    GfxFillPattern(const uint8* init);
    GfxFillPattern(uint8 val);

    bool isBlank() const;
    bool isBlack() const;

    GfxFillPattern& shiftLeft(int amount);
    GfxFillPattern& shiftRight(int amount);
    GfxFillPattern& shiftUp(int amount);
    GfxFillPattern& shiftDown(int amount);

    GfxFillPattern& operator |= (const GfxFillPattern& rhs);
    GfxFillPattern& operator |= (uint8 value);
    GfxFillPattern& operator &= (const GfxFillPattern& rhs);
    GfxFillPattern& operator &= (uint8 value);
    GfxFillPattern& operator ^= (const GfxFillPattern& rhs);
    GfxFillPattern& operator ^= (uint8 value);

    GfxFillPattern& invert();

    /** Read/write access to line i. */
    uint8& operator[](unsigned i)
        { return pat[i & 7]; }
    /** Read access to line i. */
    uint8 operator[](unsigned i) const
        { return pat[i & 7]; }
};

/** Point. This object contains a point in the X/Y plane. This is a
    plain old structure, don't hesitate to modify it directly. */
struct GfxPoint {
    int x;    ///< X coordinate.
    int y;    ///< Y coordinate.

    /** Create uninitialized(!) point. */
    GfxPoint()
        { }
    /** Create point at specified X/Y. */
    GfxPoint(int ax, int ay)
        : x(ax), y(ay) { }
    /** Compare two points for equality. */
    bool operator==(GfxPoint rhs) const
        { return x == rhs.x && y == rhs.y; }
    /** Compare two points for inequality. */
    bool operator!=(GfxPoint rhs) const
        { return x != rhs.x || y != rhs.y; }
};

/** Rectangle. An SDL_Rect or a faster implementation. Represents
    rectangles as position plus extent. This offers some useful
    primitives to work with rectangles. Otherwise, this is a plain old
    structure and you are invited to modify it directly. */
class GfxRect : public SDL_Rect {
 public:
    GfxRect();
    GfxRect(const SDL_Rect& r);
    GfxRect(int x, int y, int w, int h);
    explicit GfxRect(GfxSystemSurface sfc);

    void clipSurface(GfxSystemSurface sfc);
    void clipRect(const SDL_Rect& r);
    void include(const SDL_Rect& r);
    bool contains(const GfxRect& r) const;
    bool operator==(const SDL_Rect& rhs) const;
    bool operator!=(const SDL_Rect& rhs) const;
    GfxPoint moveTo(GfxPoint where);

    /** Check whether this rectangle has a non-zero extent. */
    bool exists() const
        { return w > 0 && h > 0; }
    /** Compute intersection. Same as clipRect(), see there. */
    void intersect(const SDL_Rect& r)
        { clipRect(r); }
    /** Check whether rectangle contains a given point. */
    bool contains(int px, int py) const
        {
            px -= x;
            py -= y;
            return px >= 0 && py >= 0 && unsigned(px) < w && unsigned(py) < h;
        }
    /** Check whether rectangle contains a given point. */
    bool contains(GfxPoint pt) const
        { return contains(pt.x, pt.y); }
    /** Move to point. \overload */
    GfxPoint moveTo(int px, int py)
        { return moveTo(GfxPoint(px, py)); }
    /** Move by relative distance. The distance is given in the
        components of a point. */
    void moveBy(GfxPoint dist)
        { x += dist.x; y += dist.y; }
    /** Move by relative distance. The distance is given by two
        integers. */
    void moveBy(int dx, int dy)
        { x += dx; y += dy; }
    /** Move by relative distance, reverse direction. */
    void moveByRev(GfxPoint dist)
        { x -= dist.x; y -= dist.y; }
    /** Move by relative distance, reverse direction. */
    void moveByRev(int dx, int dy)
        { x -= dx; y -= dy; }
    /** Grow rectangle. This enlarges the rectangle by \c dx columns
        to the left and right, and by \c dy lines at top and bottom
        (for a total growth of twice the given amount, of course). Use
        negative values to shrink. */
    void grow(int dx, int dy)
        { x -= dx; y -= dy; w += 2*dx; h += 2*dy; }
    /** Get center point of rectangle. */
    GfxPoint getCenter() const
        { return GfxPoint(x + w/2, y + h/2); }
    /** Check whether this rectangle intersects another one. */
    bool isIntersecting(GfxRect r) const
        {
            r.clipRect(*this);
            return r.exists();
        }
    /** Get bottom coordinate. */
    int getBottomY() const
        { return y + h; }
    /** Get right coordinate. */
    int getRightX() const
        { return x + w; }
};

class GfxRectSet {
    std::list<GfxRect> the_list;

    static void doSubtract(const GfxRect& r, std::list<GfxRect>& l2, std::list<GfxRect>& other);
 public:
    typedef std::list<GfxRect>::const_iterator iterator;

    GfxRectSet();
    GfxRectSet(const GfxRect& r);

    void intersect(const GfxRect& r);
    void add(const GfxRect& r);
    void remove(const GfxRect& r);
    void clear()
        { the_list.clear(); }

    bool contains(int x, int y) const;
    bool contains(const GfxPoint& r) const
        { return contains(r.x, r.y); }
    bool contains(const GfxRect& r) const;

    GfxRect getBoundingRectangle() const;

    /** Iterator interface: get iterator to first element. */
    iterator begin() const
        { return the_list.begin(); }
    /** Iterator interface: get iterator to last element. */
    iterator end() const
        { return the_list.end(); }
    /** Check whether this set is empty. */
    bool empty() const
        { return the_list.empty(); }
    /** Get number of rectangles in this rectangle set. */
    std::list<GfxRect>::size_type getNumRectangles() const
        { return the_list.size(); }
};

extern const GfxFillPattern SOLID_FILL, GRAY50_FILL, GRAY25_FILL,
    GRAY50_FILL2, LTSLASH_FILL;

std::ostream& operator<<(std::ostream&, const GfxRect&);
std::ostream& operator<<(std::ostream&, GfxPoint);

#endif
