/**
  *  \file int/exec.cc
  *  \brief Execution of BCOs
  *
  *  This provides the functions to execute user-defined subroutines
  *  (BCOs).
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/exec.h"
#include "cpluslib/assert.h"
#include "int/fun-tbl.h"
#include "int/lex.h"
#include "util/console.h"
#include "int/expr.h"

#define BCASSERT(x) ASSERT(x)

/******************************** IntFrame *******************************/

IntFrame::IntFrame(IntByteCodeObject& bco, std::size_t context_tos)
    : bco(bco), ip(0), context_tos(context_tos)
{
    for (uint32 n = 0; n < bco.getNumArgs(); ++n)
        local_names.add(bco.getArgName(n));
    bco.incUseCount();
}

IntFrame::IntFrame(const IntFrame& rhs)
    : bco(rhs.bco), ip(rhs.ip), local_names(rhs.local_names),
      local_values(rhs.local_values), context_tos(rhs.context_tos)
{
    bco.incUseCount();
}

IntFrame::~IntFrame()
{
    bco.decUseCount();
}

IntFrame*
IntFrame::clone() const
{
    return new IntFrame(*this);
}

IntExecutionContext::Exception::Exception(IntExecutionContext& ctx,
                                          IntByteCodeObject::index_t ip)
    : context_tos(ctx.context.size()), control_tos(ctx.control_stack.size()),
      value_tos(ctx.value_stack.size()), ip(ip)
{ }

/************************** IntExecutionContext **************************/

IntExecutionContext::IntExecutionContext()
    : status(ps_Stopped),
      priority(50)
{ }

IntExecutionContext::IntExecutionContext(const IntExecutionContext& rhs)
    : context(rhs.context.size()),              // just reserve space
      control_stack(rhs.control_stack.size()),  // just reserve space
      value_stack(rhs.value_stack),             // copy
      exception_stack(rhs.exception_stack),     // copy
      status(ps_Stopped),
      priority(50)
{
    for (context_t::size_type i = 0; i < context.size(); ++i)
        context[i] = rhs.context[i]->clone();
    for (control_stack_t::size_type i = 0; i < control_stack.size(); ++i)
        control_stack[i] = rhs.control_stack[i]->clone();
}

IntExecutionContext::~IntExecutionContext()
{ }

void
IntExecutionContext::pushContext(Ptr<IntContext> con)
{
    context.push_back(con);
}

void
IntExecutionContext::popContext()
{
    ASSERT(!context.empty());
    context.pop_back();
}

void
IntExecutionContext::pushFrame(Ptr<IntFrame> f)
{
    control_stack.push_back(f);
    // context.push_back(new IntFrameContext(f->vars));
}

// void
// IntExecutionContext::popFrame()
// {
//     ASSERT(!control_stack.empty());
//     control_stack.pop_back();
// }

/** Handle an exception (equivalent of `Abort'). This will either
    reflect the exception to the program's handler, or throw a C++
    exception. */
void
IntExecutionContext::doException(string_t s)
{
    while (1) {
        /* anyone catching this exception? */
        if (exception_stack.empty()) {
            throw IntException(s);
        }

        /* can we go back to that exception? */
        Exception x = exception_stack.back();
        exception_stack.pop_back();
        if (x.context_tos > context.size()
            || x.value_tos > value_stack.size()
            || x.control_tos > control_stack.size())
        {
            /* this normally cannot happen */
            // FIXME: i18n?
            console.write(LOG_ERROR, "Got stuck during exception handling. Continuing anyway...");
            continue;
        }

        /* okay, back out */
        context.erase(context.begin() + x.context_tos, context.end());
        value_stack.erase(value_stack.begin() + x.value_tos, value_stack.end());
        control_stack.erase(control_stack.begin() + x.control_tos, control_stack.end());
        break;
    }
}

/** Run one byte-code. */
void
IntExecutionContext::runOnce()
{
    ASSERT(!control_stack.empty());
    Ptr<IntFrame> f = control_stack.back();

    uint32 code = f->bco.getCode(f->ip++);
//     std::cout << (f->ip-1) << ": `" << code << "' [" << int32(getOpcodeFromCode(code))
//               << "," << getArgFromCode(code)
//               << "], value stack size = " << value_stack.size()
//               << "\n";
    switch(getOpcodeFromCode(code)) {
     case op_CallUnary: {
         /* r:ARG -> (op_CallUnary, index into unary_table) -> r:result */
         BCASSERT(value_stack.size() > 0);
         pushV(unary_table[getArgFromCode(code)](popV()));
         break;
     }

     case op_CallBinary: {
         /* r:A1:A2 -> (op_CallBinary, index into binary_table) -> r:result */
         BCASSERT(value_stack.size() > 1);
         IntObject a2 = popV(), a1 = popV();
         pushV(binary_table[getArgFromCode(code)](a1, a2));
         break;
     }

     case op_CallTernary: {
         /* r:A1:A2:A3 -> (op_CallTernary, index into ternary_table) -> r:result */
         BCASSERT(value_stack.size() > 2);
         IntObject a3 = popV(), a2 = popV(), a1 = popV();
         pushV(ternary_table[getArgFromCode(code)](a1, a2, a3));
         break;
     }

     case op_CallVA:
        ASSERT(!"implemented");

     case op_InvokeUser: {
         ASSERT(!value_stack.empty());
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject nargs = popV();
         IntSubroutine* sub = IntSubroutine::getSubrByName(name);
         if (!nargs.isInt())
             doException("Bytecode error: op_InvokeUser");
         if (!sub) {
             doException("No subroutine '" + name + "' known");
         } else {
             sub->run(nargs.getIntVal(), name, *this);
         }
         break;
     }

     case op_InvokeContext: {
         ASSERT(!value_stack.empty());
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntContext::handle_t v = context.back()->lookupSub(name);
         IntObject nargs = popV();
         if (!nargs.isInt())
             doException("Bytecode error: op_InvokeContext");
         if (v < 0) {
             doException("Command '" + name + "' not known in this context");
         } else {
             context.back()->call(v, nargs.getIntVal(), *this);
         }
         break;
     }

     case op_PushNull:
        /* r -> (op_PushNull) -> r:NULL */
        pushV(IntObject());
        break;

     case op_PushBool:
        /* r -> (op_Push{Bool,Int}, V) -> r:V */
        pushV(IntObject(getArgFromCode(code) != 0));
        break;

     case op_PushInt:
        pushV(IntObject(int32(getArgFromCode(code))));
        break;

     case op_PushFloat:
        /* r -> (op_Push{Float,String}, index) -> r:V */
        pushV(IntObject(f->bco.getFloat(getArgFromCode(code))));
        break;

     case op_PushString:
        pushV(IntObject(f->bco.getStr(getArgFromCode(code))));
        break;

     case op_PushIntHi:
        /* r:V -> (op_PushIntHi, HI) -> r:(HI<<24 + V) */
        BCASSERT(value_stack.size() > 0);
        BCASSERT(value_stack.back().isInt());
        value_stack.back() = int32(value_stack.back().getIntVal() | (getArgFromCode(code) << 24));
        break;

     case op_Pop: {
         /* r:A1:...:An -> (op_Pop, n) -> r */
         uint32 count = getArgFromCode(code);
         BCASSERT(value_stack.size() >= count);
         value_stack.erase(value_stack.end() - count, value_stack.end());
         break;
     }

     case op_Dup: {
         uint32 index = getArgFromCode(code);
         BCASSERT(value_stack.size() > index);
         pushV(value_stack[value_stack.size()-index-1]);
         break;
     }

     case op_DupC: {
         uint32 index = getArgFromCode(code);
         BCASSERT(context.size() > index);
         pushContext(context[context.size()-index-1]);
         break;
     }

     case op_QualLookup: {
         /* r:CONTEXT -> (op_QualLookup, index into string table (name)) -> r:VALUE */
         // FIXME: code duplication from IntQualLookup::eval
         BCASSERT(value_stack.size() > 0);
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject c = popV();
         if (c.getType() != res_Context)
             throw IntException("Left operand of `.' is not a context");
         IntContext::handle_t handle = c.getContextRef().lookup(name);
         if (handle < 0)
             throw IntException("Unknown field `" + name + "'");
         pushV(c.getContextRef().read(handle));
         break;
     }

     case op_QualAssign: {
        /* r:VALUE:CONTEXT -> (op_QualAssign, -"-) -> r */
         BCASSERT(value_stack.size() > 1);
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject c = popV(), v = popV();
         if (c.getType() != res_Context)
             throw IntException("Left operand of `.' is not a context");
         IntContext::handle_t handle = c.getContextRef().lookup(name);
         if (handle < 0)
             throw IntException("Unknown field `" + name + "'");
         c.getContextRef().store(handle, v);
         break;
     }

     case op_UnqualLookup: {
         /* r -> (op_UnqualLookup, -"-) -> r:VALUE */
         string_t name = f->bco.getStr(getArgFromCode(code));
         std::size_t cx_index = context.size();
         for (std::size_t i = control_stack.size(); i > 0; --i) {
             /* first, look in all contexts above it */
             while (cx_index > control_stack[i-1]->context_tos) {
                 IntContext::handle_t h = context[--cx_index]->lookup(name);
                 if (h >= 0) {
                     pushV(context[cx_index]->read(h));
                     return;
                 }
             }
             IntVariableNames::index_t h = control_stack[i-1]->local_names.lookup(name);
             if (h >= 0) {
                 pushV(control_stack[i-1]->local_values.read(h));
                 return;
             }
         }
         while (cx_index > 0) {
             IntContext::handle_t h = context[--cx_index]->lookup(name);
             if (h >= 0) {
                 pushV(context[cx_index]->read(h));
                 return;
             }
         }
         throw IntException("Unknown identifier `" + name + "'");
     }

     case op_UnqualAssign: {
        /* r:VALUE -> (op_UnqualAssign, -"-) -> r */
         BCASSERT(value_stack.size() > 0);
         string_t name = f->bco.getStr(getArgFromCode(code));
         std::size_t cx_index = context.size();
         for (std::size_t i = control_stack.size(); i > 0; --i) {
             /* first, look in all contexts above it */
             while (cx_index > control_stack[i-1]->context_tos) {
                 IntContext::handle_t h = context[--cx_index]->lookup(name);
                 if (h >= 0) {
                     context[cx_index]->store(h, popV());
                     return;
                 }
             }
             IntVariableNames::index_t h = control_stack[i-1]->local_names.lookup(name);
             if (h >= 0) {
                 control_stack[i-1]->local_values.store(h, popV());
                 return;
             }
         }
         while (cx_index > 0) {
             IntContext::handle_t h = context[--cx_index]->lookup(name);
             if (h >= 0) {
                 context[cx_index]->store(h, popV());
                 return;
             }
         }
         throw IntException("Unknown identifier `" + name + "'");
     }

     case op_Apply:
        /* r:A1:...:An:FUN -> (op_Apply, n) -> r:VALUE */
        FIXME("unimplemented: APPLY");
     case op_ApplyAssign:
        /* r:VALUE:A1:...:An:FUN -> (op_ApplyAssign, n) -> r */
        FIXME("unimplemented: APPLYASSIGN");

    /* keep this order. Basic idea: after subtracting op_DontJump,
       we get a 4-bit field:
         bit 0:    jump if value is true
         bit 1:    jump if value is false
         bit 2:    jump if value is empty
         bit 3:    pop TOS after successful jump */
    /* r:VALUE -> (op_JumpXX, address) -> r:VALUE */
    /* r:VALUE -> (op_JumpXXPop, address) -> r */
     case op_DontJump:
        break;
     case op_Jump: {
         // this is a special case because it doesn't need a stack
         // element. Not special-casing it would bomb out with an
         // empty stack otherwise
         f->ip = f->bco.getLabel(getArgFromCode(code));
         break;
     }
     case op_JumpT:                   /* if true */
     case op_JumpF:                   /* if false */
     case op_JumpTF:
     case op_JumpE:                   /* if empty */
     case op_JumpTE:
     case op_JumpFE:
     case op_DontJumpPop:
     case op_JumpTPop:                /* if true */
     case op_JumpFPop:                /* if false */
     case op_JumpTFPop:
     case op_JumpEPop:                /* if empty */
     case op_JumpTEPop:
     case op_JumpFEPop:
     case op_JumpPop: {
         int    cond = getOpcodeFromCode(code) - op_DontJump;
         uint32 addr = getArgFromCode(code);
         IntObject tester = (cond & 8) ? popV() : value_stack.back();
         if (!tester.isScalar())
             throw IntException("Tested value is not scalar");

         int res = tester.getBoolVal();
         if ((res == -1 && (cond & 4))
             || (res == 0 && (cond & 2))
             || (res == 1 && (cond & 1)))
         {
             f->ip = f->bco.getLabel(addr);
         }
         break;
     }

     case op_EvalStr: {
         /* r -> (op_EvalStr, index into string table) -> r */
         doEval(f->bco.getStr(getArgFromCode(code)));
         break;
     }
     case op_SetLine: {
         /* r -> (op_SetLine, line nr) -> r */
         break;
     }

     case op_PushContext: {
         /* r:CONTEXT -> r */
         BCASSERT(value_stack.size() > 0);
         IntObject o = popV();
         if (!o.isContext())
             throw IntException("Argument is not a context");
         pushContext(o.getContextVal());
         break;
     }

     case op_PopContext: {
        uint32 i = getArgFromCode(code);
        BCASSERT(context.size() > i);
        while (i--)
            popContext();
        break;
     }

     case op_CurrentContext: {
         BCASSERT(context.size() > 0);
         pushV(IntObject(context.back()));
         break;
     }

     case op_FirstContext: {
         /* r:ARRAY -> r */
         /* go to first element and op_PushContext, push T or F */
         BCASSERT(value_stack.size() > 0);
         IntObject o = popV();
         IntArray* a;
         if (!o.isFunction() || !(a = dynamic_cast<IntArray*>(&o.getFunctionRef())))
             throw IntException("Argument is not an array");
         if (Ptr<IntContext> first = a->getFirst()) {
             pushContext(first);
             pushV(true);
         } else {
             pushV(false);
         }
         break;
     }

     case op_NextContext: {
         /* go to next element if possible, push T or F */
         BCASSERT(context.size() > 0);
         pushV(context.back()->next());
         break;
     }

     /* start a try block, on exception, push it and jump to address */
     case op_Try: {
         exception_stack.push_back(Exception(*this,
                                             f->bco.getLabel(getArgFromCode(code))));
         break;
     }
     case op_EndTry: {
        BCASSERT(exception_stack.size() > 0);
        // FIXME: should verify that all other pointers are ok?
        exception_stack.pop_back();
        break;
     }
     case op_Throw: {
        BCASSERT(value_stack.size() > 0);
        IntObject n = popV();
        if (!n.convertToString())
            doException("Invalid exception object");
        else
            doException(n.getStringVal());
        break;
     }

     case op_DefineSub: {
        // r:NAME -> (op_DefineSub, subr_id) -> r
        BCASSERT(value_stack.size() > 0);
        IntObject n = popV();
        if (!n.convertToString() || n.getStringVal().length() == 0)
            throw IntException("Invalid subroutine name");

        IntSubroutine* sub = IntSubroutine::getSubrByIndex(getArgFromCode(code));
        BCASSERT(sub);
        sub->setSubrName(strUCase(n.getStringVal()));
        break;
     }

     case op_DimLocal: {
         BCASSERT(value_stack.size() > 0);
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject n = popV();
         if (f->local_names.lookup(name) < 0)
             f->local_values.store(f->local_names.add(name), n);
         break;
     }

     case op_DimStatic: {
         BCASSERT(value_stack.size() > 0);
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject n = popV();
         if (control_stack.front()->local_names.lookup(name) < 0)
             control_stack.front()->local_values.store(control_stack.front()->local_names.add(name), n);
         break;
     }

     case op_DimShared: {
         BCASSERT(value_stack.size() > 0);
         string_t name = f->bco.getStr(getArgFromCode(code));
         IntObject n = popV();
         if (IntVariableNames::global_names.lookup(name) < 0)
             IntVariableValues::global_values.store(IntVariableNames::global_names.add(name), n);
         break;
     }

     case op_Return: {
        doReturn();
        break;
     }

     case op_SetStatus: {
        BCASSERT(getArgFromCode(code) >= 0 && getArgFromCode(code) <= ps_LAST);
        status = Status(getArgFromCode(code));
        break;
     }

     default:
        BCASSERT(0);
    }
}

void
IntExecutionContext::doReturn()
{
    Ptr<IntFrame> f = control_stack.back();

    /* remove all exception handlers of that frame */
    while (!exception_stack.empty() && exception_stack.back().control_tos >= control_stack.size())
        exception_stack.pop_back();

    /* remove all contexts */
    while (!context.empty() && context.size() >= f->context_tos) {
        Ptr<IntContext> ctx = context.back();
        popContext();
//         if (IntFrameContext* con = dynamic_cast<IntFrameContext*>(ctx.ptr())) {
//             ASSERT(&con->getVariableRef() == &f->vars);
//             break;
//         }
    }

    /* finally, remove the frame */
    control_stack.pop_back();

    /* terminate the process */
    if (control_stack.empty())
        status = ps_Terminated;
}

void
IntExecutionContext::run()
{
    status = ps_Running;
    while (status == ps_Running) {
        try {
            IntFrame* f = control_stack.back();
            if (f->ip >= f->bco.getCodeSize())
                doReturn();
            else
                runOnce();
        }
        catch(IntException& e) {
            doException(string_t(e.what()));
        }
    }
}

void
IntExecutionContext::doEval(string_t s)
{
    // FIXME: this function has *lots* of problems and optimisation
    // potential.
    IntByteCodeObject& tmp_bco = IntByteCodeObject::allocBCO();
    IntLexer lex(s);
    if (lex.lex() == tt_Id) {
        /* Look whether it is a subroutine. */
        string_t name = lex.getString();
        for (std::size_t i = context.size(); i > 0; --i) {
            IntContext::handle_t n = context[i-1]->lookupSub(name);
            if (n >= 0) {
                /* It is a program-defined subroutine. */
                int nargs = parseArgs(lex, tmp_bco);
                tmp_bco.pushInt(nargs);
                if (i == context.size()) {
                    tmp_bco.add(mkCode(op_InvokeContext, tmp_bco.allocStr(name)));
                } else {
                    tmp_bco.add(mkCode(op_DupC, context.size() - i));
                    tmp_bco.add(mkCode(op_InvokeContext, tmp_bco.allocStr(name)));
                    tmp_bco.add(mkCode(op_PopContext, 1));
                }
                return;
            }
        }

        /* Look whether there's a user-defined or global subroutine. */
        IntSubroutine* subr = IntSubroutine::getSubrByName(name);
        if (subr) {
            int nargs = parseArgs(lex, tmp_bco);
            tmp_bco.pushInt(nargs);
            tmp_bco.add(mkCode(op_InvokeUser, tmp_bco.allocStr(name)));
            tmp_bco.run(0, string_t("IntExecutionContext::doEval"), *this);
            return;
        }
    }

    /* When we're here, it must be an expression */
    IntExprParser xp(lex);
    xp.parseExpr();
    xp.pop()->compileV(tmp_bco);
    if (lex.getType() != tt_Null)
        doException("Expression incorrectly terminated");
    pushFrame(new IntFrame(tmp_bco, control_stack.size()));
}

int
IntExecutionContext::parseArgs(IntLexer& lex, IntByteCodeObject& bco)
{
    int num_args = 0;
    if (lex.lex() != tt_Null) {
        while (1) {
            IntExprParser xp(lex);
            xp.parseExpr();
            xp.pop()->compileV(bco);
            ++num_args;
            if (lex.getType() == tt_Null)
                break;
            if (lex.getType() != tt_Comma)
                doException("Expected `,' after expression");
            lex.lex();
        }
    }
    return num_args;
}

/***************************** IntProcessList ****************************/

IntProcessList::IntProcessList()
{ }

IntProcessList::~IntProcessList()
{ }

void
IntProcessList::runRunnableProcesses()
{
    for (process_list_t::iterator i = list.begin(); i != list.end(); ++i) {
        if ((*i)->getStatus() == IntExecutionContext::ps_Runnable) {
            try {
                (*i)->run();
            }
            catch(IntException& e) {
                (*i)->setStatus(IntExecutionContext::ps_Finished);
                console.write(LOG_ERROR, (*i)->getName() + ": " + e.what());
            }
        }
    }
}

void
IntProcessList::runAllProcesses()
{
    for (process_list_t::iterator i = list.begin(); i != list.end(); ++i)
        if ((*i)->getStatus() == IntExecutionContext::ps_Stopped)
            (*i)->setStatus(IntExecutionContext::ps_Runnable);
    runRunnableProcesses();
}

void
IntProcessList::addProcess(Ptr<IntExecutionContext> process)
{
    const int pri = process->getPriority();

    process_list_t::iterator j = list.begin();
    while (j != list.end() && (*j)->getPriority() > pri)
        ++j;
    list.insert(j, process);
}

void
IntProcessList::killTerminatedProcesses()
{
    process_list_t::iterator i = list.begin();
    while (i != list.end()) {
        if ((*i)->getStatus() == IntExecutionContext::ps_Terminated || (*i)->getStatus() == IntExecutionContext::ps_Finished)
            i = list.erase(i);
        else
            ++i;
    }
}

void
IntProcessList::reschedule(const Ptr<IntExecutionContext> process, const int pri)
{
    if (pri == process->getPriority())
        return;

    process_list_t::iterator i = list.begin();
    while (i != list.end() && *i != process)
        ++i;
    ASSERT(i != list.end());
    list.erase(i);

    (*i)->setPriority(pri);

    process_list_t::iterator j = list.begin();
    while (j != list.end() && (*j)->getPriority() > pri)
        ++j;
    list.insert(j, process);
}

/*************************** Interface Helpers ***************************/

int32
IF_intValue(const IntObject& o, const char* name)
{
    IntObject c (o);

    if (c.convertToInt())
        // FIXME: unsure about the following one, but 1.x does it, too.
        // This way, "Speed$ := 9.5" and "SetSpeed 9.5" work, and set the
        // speed to 9.
        return c.getIntVal();
    else
        throw IntException("Value for `" + string_t(name) + "' must be integer");
}

bool
IF_boolValue(const IntObject& o, const char* name)
{
    // FIXME: PCC 1.x has pretty messy to-bool conversions:
    // - in assignments, numbers convert to bool, strings don't; all
    //   floating point values below 1 convert to false.
    // - in function/sub calls and core commands, all types convert
    //   to bool, fp values below EPSILON convert to false.
    // Right now, we always use the latter convention.
    int i = o.getBoolVal();
    if (i >= 0)
        return i;
    else
        throw IntException("Value for `" + string_t(name) + "' must be boolean");
}

double
IF_floatValue(const IntObject& o, const char* name)
{
    if (o.isInt() || o.isBool())
        return o.getIntVal();
    else if (o.isFloat())
        return o.getFloatVal();
    else
        throw IntException("Value for `" + string_t(name) + "' must be numeric");
}

string_t
IF_stringValue(IntObject o, const char* name)
{
    if (!o.convertToString())
        throw IntException("Value for `" + string_t(name) + "' must be string");
    return o.getStringVal();
}

bool
IF_argCheck(int got_args, int min_args, int max_args,
            const char* name, IntExecutionContext& xc)
{
    if (got_args < min_args)
        throw IntException("Too few arguments to `" + string_t(name) + "'");
    if (got_args > max_args)
        throw IntException("Too many arguments to `" + string_t(name) + "'");

    ASSERT(xc.getValueStack().size() >= got_args);

    // validate that all mandatory arguments are non-empty. If one is
    // empty, the command is skipped.
    for (int i = min_args+1; i <= got_args; ++i)
        if (xc.peekV(got_args - i).isEmpty())
            return false;
    return false;
}
