#!/usr/bin/perl -w
#
#  Interface Compiler for PCC II
#
#  This compiles the description of a CCScript context in a higher-level
#  syntax into actual C++ code. The syntax is inspired by the one used
#  in PCC 1.x, but not identical.
#
#  Usage: ifcomp input output_without_h_or_cc
#  e.g. "ifcomp int/if-planet.if int/if-planet" will generate
#  "int/if-planet.cc" and "int/if-planet.h".
#
#  Input syntax:
#     introtext
#     %%
#     interface definition
#  The introtext is copied into the .cc file.
#
#  Commands allowed in the interface definition:
#
#  - 'class NAME;'                 set name of the class we're generating.
#                                  Mandatory, should be first.
#  - 'VSPEC TYPE NAME;'            define a class member. VSPEC is public,
#                                  private or protected.
#  - 'VSPEC TYPE NAME(ARGS) { }'   define a member function.
#  - 'using VSPEC TYPE MEMBER "file.h";'  define a member of class type,
#                                  and bring its members into scope. Type
#                                  must be a class, pointer-to-class or
#                                  reference-to-class. The class definition
#                                  must be in "file.h".
#  - 'N:T, N:T {READER}{WRITER}    define one or more properties which
#                                  have the specified reader/writer code.
#                                  The property is read-only if no writer
#                                  is specified
#  - 'sub NAME(args) { }'          define a command specific to this context.
#                                  args are comma-separated arguments; each
#                                  can be
#                                    ['optional'] name [:type] [=default]
#                                  -- optional: this and all following args
#                                     are optional
#                                  -- name: name of this arg on the C++ side
#                                  -- type: type (int/bool/...); if not
#                                     specified, all types do (IntObject)
#                                  -- default: must be specified for typed
#                                     optional parameters; untyped optional
#                                     params come out EMPTY.
#                                  Caller of your code makes sure that all
#                                  mandatory args are there and non-empty,
#                                  and that all types are ok.
#
#  C++ code can appear in the following places:
#  - readercode and writercode of a property
#  - implementation of a member function
#  - code of a subroutine
#  - [in the introtext]
#
#  All C++ code except for the introtext is transformed by the generator:
#  - '$foo' translates into an expression which is 'true' when we're
#    currently reading/writing property 'foo'. Valid only in reader/
#    writercode.
#  - 'FAIL' causes an assignment error. Valid only in writercode.
#  - 'VALUE' is the value to be assigned. Valid only in writercode
#  - 'VALUE_RAW' is the raw (IntObject) value to be assigned. Valid
#    only in writercode.
#  - 'ERROR(xx)' causes an error message specified by 'xx' (a string).
#    Valid in readercode, writercode and subroutines.
#  - unqualified names are translated to member accesses for 'using'd
#    members. For example, after 'using GPlanet* p "game/planet.h"',
#    'getId()' translates into 'p->getId()'.
#
#  TODO:
#  - change syntax from Pascal'ish "name : type" to C'ish "type name"?
#  - generate documentation from .if file?
#  - are there more efficient renderings?
#

use strict;
use vars qw{ $arg_input $arg_output_base };
use vars qw{ @c_output @h_output };

use vars qw{ $g_class_name @g_class_body $g_class_visibility };
use vars qw{ %g_name_data %g_sub_data };
use vars qw{ %g_reader_codes %g_writer_codes };
use vars qw{ %g_source_cache };
use vars qw{ %g_seen_functions };

use vars qw{ %g_type_accessor_code %g_type_cxx_equivalent };

use vars qw{ %g_unqual_translation };

# Variables for Lexer
use vars qw{ $lex_line $lex_token };

%g_type_cxx_equivalent = ("int" => "int32",
                  "bool" => "bool",
                  "float" => "double",
                  "string" => "string_t",
                  "context" => "Ptr<IntContext>",
                  "function" => "Ptr<IntFunction>");
%g_type_accessor_code = ("int" => "IF_intValue(_value, NAME)",
                         "bool" => "IF_boolValue(_value, NAME)",
                         "float" => "IF_floatValue(_value, NAME)",
                         "string" => "IF_stringValue(_value, NAME)");
while (@ARGV) {
    if (!defined $arg_input) {
        $arg_input = shift @ARGV;
    } elsif (!defined $arg_output_base) {
        $arg_output_base = shift @ARGV;
    } else {
        die "invalid usage.\n";
    }
}
if (!defined $arg_output_base) {
    die "invalid usage.\n";
}

push @c_output, "/* Generated file, do not edit */\n";
push @h_output, "/* Generated file, do not edit */\n";
my $macro_stem = uc $arg_output_base;
$macro_stem =~ s/\W+/_/g;
push @h_output,
    "#ifndef PCC_V2_${macro_stem}_H\n",
    "#define PCC_V2_${macro_stem}_H\n",
    "\n",
    "#include \"int/data.h\"\n\n";

open INPUT, "< $arg_input" or die "$arg_input: $!\n";
while (<INPUT>) {
    last if /^%%/;
    push @c_output, $_;
}
foreach my $header ('int/data.h', 'int/lex.h', 'int/exec.h', 'util/bsearch.h', "$arg_output_base.h") {
    push @c_output, "#include \"$header\"\n"
        unless grep /^#include\s*\"$header\"/, @c_output;
}

$g_class_visibility = 'private';
# expand_add_translation ('FAIL', 'return false');       # used for assignments
expand_add_translation ('ERROR', 'throw IntException');  # subs will (?) override this
lex_eof(0);
while (1) {
    if ($lex_token eq '%%') {
        last;
    } elsif ($lex_token eq '**EOF**') {
        last;
    } elsif ($lex_token eq 'class') {
        read_class ();
    } elsif ($lex_token eq 'using') {
        read_using ();
    } elsif ($lex_token eq 'sub') {
        read_sub ();
    } elsif ($lex_token eq 'private' || $lex_token eq 'protected' || $lex_token eq 'public') {
        read_member ();
    } else {
        read_properties ();
    }
}
close INPUT;

if (!defined $g_class_name) {
    die "need class name (`class NAME;' statement";
}

emit_name_array_and_number_objects ();
emit_lookup_function ();
generate_codes ();
emit_reader ();
emit_writer ();
emit_subs ();
emit_default_functions ();

# foreach (sort keys %g_unqual_translation) {
#     print "$_ => $g_unqual_translation{$_}\n";
# }

push @h_output, "class $g_class_name : public IntContext {\n";
push @h_output, @g_class_body;
push @h_output, "};\n\n#endif\n";

do_output ("$arg_output_base.cc", @c_output);
do_output ("$arg_output_base.h", @h_output);

exit 0;

############################# Reader functions ############################

# read a 'class' statement:
#   class NAME ;
sub read_class {
    if (defined $g_class_name) { die "we already had a `class' statement"; }

    $g_class_name = lex(0);
    if ($g_class_name =~ /[^0-9a-zA-Z_:]/) { die "invalid class name"; }

    lex (0);
    lex_expect (';');
}

# read a member definition
sub read_member {
    if (!defined $g_class_name) { die "no `class' statement"; }

    # first, emit visibility specifier
    class_set_visibility ($lex_token);

    # read member name
    my @list;
    lex(0);
    while ($lex_token ne '(' && $lex_token ne '[' && $lex_token ne ';') {
        if ($lex_token =~ /\S/) { push @list, $lex_token; } else { push @list, ' '; }
        lex(1);
    }
    while (@list && $list[-1] =~ /^\s+$/) { pop @list; }
    if (!@list) { die "did not understand this member definition"; }

    my $name = pop @list;
    if (@list && $list[-1] eq '::') {
        die "invalid class definition"
            if @list < 2 || $list[-2] ne $g_class_name;
        pop @list;
        pop @list;
    }
    expand_add_translation ($name, $name);
    if ($lex_token eq '(') {
        # it is a member function definition
        my $proto = $lex_token;
        lex(1);
        while ($lex_token ne '{' && $lex_token ne ';' && $lex_token ne ':') {
            $proto .= $lex_token;
            lex(1);
        }
        $proto =~ s/\s+$//;
        class_add ('    ', join('', @list, $name, $proto, ";\n"));

        if ($lex_token eq '{' || $lex_token eq ':') {
            my $code = expand_code (lex_function_body (), '', 0);
            push @c_output, join('', "\n", @list, $g_class_name, '::', $name, $proto, "\n", $code, "\n");
        }
        $g_seen_functions{$name} = 1;
    } else {
        # it is a member variable
        my $rest = '';
        while ($lex_token ne ';') { $rest .= $lex_token; lex(1); }
        class_add ('    ', join('', @list, $name, $rest, "\n"));
    }
    lex_eof(0);
}

# Properties
sub read_properties {
    my @name_list;
    my @type_list;

    # read name/type pairs
    while (1) {
        $lex_token =~ /^\w/ or die "invalid property name `$lex_token'";
        my $real_name = uc $lex_token;
        push @name_list, $real_name;
        lex(0);
        lex_expect(':');

        my $real_type = read_type ();
        if (exists $g_name_data{$real_name}) { die "'$real_name' already declared'" }
        $g_name_data{$real_name} = { type => $real_type };

        if ($lex_token eq '{') { last }
        if ($lex_token ne ',') { die "expected ',' or '{'"; }
    }

    my $reader = lex_function_body('        ');
    foreach (@name_list) { $g_name_data{$_}{reader_code} = $reader }
    if (lex_eof(0) eq '{') {
        my $writer = lex_function_body('        ');
        foreach (@name_list) { $g_name_data{$_}{writer_code} = $writer }
        lex_eof(0);
    }
}

sub read_type {
    my $type = ($lex_token =~ /^str(ing)?$/i
                ? 'string'
                : $lex_token =~ /^int(eger)?$/i
                ? 'int'
                : $lex_token =~ /^float$/i
                ? 'float'
                : $lex_token =~ /^bool(ean)?$/i
                ? 'bool'
                : $lex_token =~ /^con(text)?$/i
                ? 'context'
                : $lex_token =~ /^fun(c(tion)?)?$/i
                ? 'function'
                : die "invalid type");
    lex(0);
    $type;
}

# read a 'using' statement
#  using class_name [*|&] member_name "file" ;
sub read_using {
    lex (0);
    my $visibility = 'private';
    if ($lex_token eq 'public' || $lex_token eq 'private' || $lex_token eq 'protected') {
        $visibility = $lex_token;
        lex (0);
    }
    my $class_name = $lex_token;
    my $operator = lex (0);
    my $member;
    if ($operator ne '*' && $operator ne '&') {
        $member = $operator;
        $operator = '';
    } else {
        $member = lex (0);
    }
    my $file = lex (0);
    lex (0);
    if ($lex_token ne ';') { die "expected ';', got '$lex_token'" }

    # validate
    $class_name !~ /\W/ or die "invalid class name `$class_name'";
    $member !~ /\W/ or die "invalid member name `$member'";
    $file =~ /^\"/ or die "invalid file name `$file'";

    # add appropriate includes
    if ($operator eq '') {
        push @h_output, "#include $file\n"
            unless grep /^\#include\s*\Q$file/m, @h_output;
    } else {
        push @h_output, "class $class_name;\n"
            unless grep /^class\s+\Q$class_name;/m, @h_output;
        push @c_output, "#include $file\n"
            unless grep /^\#include\s*\Q$file/m, @c_output;
    }

    # load file
    $file =~ s/\"//g;
    if (!exists $g_source_cache{$file}) {
        open CLASS_FILE, "< $file" or die "$file: $!";
        $g_source_cache{$file} = join '', <CLASS_FILE>;
        close CLASS_FILE;
    }

    # fetch class
    my $source = $g_source_cache{$file};
    $source =~ /^(class|struct)\s+\Q$class_name\E\s*[a-zA-Z0-9_:<>\s,]*\{/mg
        or die "no `class $class_name' in `$file'";
    my $reading = ($1 eq 'struct');

    # add member to class
    class_set_visibility ($visibility);
    class_add ("    $class_name$operator $member;\n");

    # now, we can scan the source for interesting things
    # FIXME: this is essentially code duplication from read_class ()
    my $translation = ($operator eq '*' ? "$member->" : "$member.");
    lex_string (\$source, 0);
    while (defined($lex_token) && $lex_token ne '}') {
        my @list = ($lex_token);
        while (defined($lex_token) && $lex_token ne '(' && $lex_token ne '[' && $lex_token ne ';' && $lex_token ne '{' && $lex_token ne '}' && $lex_token ne ':') {
            push @list, $lex_token;
            lex_string (\$source, 0);
        }
        if (!@list) { die "did not understand this definition"; }
        if ($lex_token eq ':') {
            if ($list[-1] =~ /^p(r(ivate|otected)|ublic)$/) {
                $reading = ($list[-1] eq 'public');
            }
        } else {
            # $list[-1] now is the member name. Translate it.
            if ($reading && $list[-1] !~ /\W/ # weed out operators and reserved words
                && $list[-1] !~ /^(enum|class|struct|union|if|for|while|do|switch|const|volatile|operator)$/
                && $list[-1] ne $class_name)
            {
                expand_add_translation ($list[-1], $translation . $list[-1])
            }

            # Scan until end of the member definition.
            while (defined($lex_token) && $lex_token ne '{' && $lex_token ne '}' && $lex_token ne ';') {
                lex_string (\$source, 0);
            }

            if ($lex_token eq '{') {
                my $brace = 0;
                while (defined($lex_token)) {
                    if ($lex_token eq '{') {
                        ++$brace;
                    } elsif ($lex_token eq '}') {
                        --$brace;
                        last if $brace == 0;
                    }
                    lex_string (\$source, 0);
                }
            }
        }
        lex_string (\$source, 0);
    }
    lex (0);
}

sub read_sub {
    # here, we're looking at "sub".
    lex(0);

    my $sub_name = $lex_token; lex(0);
    my @arg_names;
    my @arg_types;
    my @arg_defaults;
    my $num_mandatory = 0;
    my $seen_optional = 0;

    # read argument names
    if ($lex_token eq '(') {
        lex(0);
        if ($lex_token eq ')') {
            # nothing
            lex(0);
        } else {
            # we're looking at the first word
            while (1) {
                my ($name, $type, $default);
                if ($lex_token eq 'optional') {
                    die "'optional' already seen" if $seen_optional;
                    $seen_optional = 1;
                    lex(0);
                }
                $name = $lex_token;
                lex(0);
                if ($lex_token eq ':') {
                    lex(0);
                    $type = read_type ();
                }
                if ($lex_token eq '=') {
                    lex(0);
                    $default = lex_expr ();
                }
                # got one, validate it
                if ($name =~ /\W/) { die "invalid name `$name'"; }
                if (defined $type && !exists $g_type_accessor_code{$type}) { die "invalid type for `$name'"; }
                if (defined $type && $seen_optional && !defined $default) { die "need default for typed optional parameter `$name'"; }
                if (defined $default && !$seen_optional) { warn "got default value for mandatory parameter `$name'?"; }
                if (grep {$_ eq $name} @arg_names) { die "`$name' already in use here"; }

                # valid. Add it.
                push @arg_names, $name;
                push @arg_types, $type; # defined $type ? $type : '';
                push @arg_defaults, $default;
                ++$num_mandatory unless $seen_optional;

                if ($lex_token eq ')') { last; }
                if ($lex_token ne ',') { die "expected ',' or ')', got '$lex_token'" }
                lex(0);
            }
            lex(0);             # skip the ')'
        }
    }

    # Here, we must be reading the body
    if ($lex_token ne '{') { die "expected '{', got '$lex_token'" }
    my $body = lex_function_body ();

    # Finally, validate the name ...
    if ($sub_name =~ /\W/) { die "invalid sub name `$sub_name'" }
    if (exists $g_sub_data{$sub_name}) { die "we already did sub `$sub_name'" }

    # Now, register the function
    $g_sub_data{uc $sub_name} = {
        body => $body,
        proper_name => $sub_name,
        num_mandatory_args => $num_mandatory,
        arg_names => \@arg_names,
        arg_types => \@arg_types,
        arg_defaults => \@arg_defaults
    };

    lex_eof (0);
}

######################## Class Prototype Generator ########################

sub class_add {
    push @g_class_body, @_;
}

sub class_set_visibility {
    my $vis = shift;
    if ($g_class_visibility ne $vis) {
        push @g_class_body, " $vis:\n";
        $g_class_visibility = $vis;
    }
}

############################ Output Generation ############################

# Generate the "names" array while numbering objects
sub emit_name_array_and_number_objects {
    my $counter;
    my @properties = sort keys %g_name_data;
    if (@properties) {
        push @c_output, "\nstatic const char* const property_names[] = {\n";
        foreach (@properties) {
            push @c_output, "    \"$_\",\n";
            $g_name_data{$_}{number} = $counter++;
        }
        push @c_output, "};\n";
    }

    my @subs = sort keys %g_sub_data;
    $counter = 0;
    if (@subs) {
        push @c_output, "\nstatic const char* const sub_names[] = {\n";
        foreach (@subs) {
            push @c_output, "    \"$_\",\n";
            $g_sub_data{$_}{number} = $counter++;
        }
        push @c_output, "};\n";
    }
}

sub emit_lookup_function {
    class_set_visibility ('public');
    class_add ("    handle_t lookup(string_t name);\n");
    class_add ("    handle_t lookupSub(string_t name);\n");
    push @c_output, ("\n",
                     "// Look up a property and say whether we know it.\n",
                     "${g_class_name}::handle_t\n",
                         "${g_class_name}::lookup(string_t name)\n",
                     "{\n",
                     (keys %g_name_data
                      ? "    return bin_search(&property_names[0], countof(property_names), IntNameLookupComparator(name));\n"
                      : "    return -1;\n"),
                     "}\n");
    push @c_output, ("\n",
                     "// Look up a subroutine name and say whether we know it.\n",
                     "${g_class_name}::handle_t\n",
                         "${g_class_name}::lookupSub(string_t name)\n",
                     "{\n",
                     (keys %g_sub_data
                      ? "    return bin_search(&sub_names[0], countof(sub_names), IntNameLookupComparator(name));\n"
                      : "    return -1;\n"),
                     "}\n");
}

sub emit_reader {
    class_set_visibility('public');
    class_add ("    IntObject read(handle_t _number);\n");
    push @c_output, ("\n",
                     "// Given a property index, return its value.\n",
                     "IntObject\n",
                     "${g_class_name}::read(handle_t _number)\n",
                     "{\n",
                     "    switch(_number) {");
    foreach (keys %g_reader_codes) {
        foreach (@{$g_reader_codes{$_}}) {
            push @c_output, "\n     case $_:";
        }
        push @c_output, $_, "\n        break;";
    }
    push @c_output, ("\n    }\n",
                     "    return IntObject();\n",
                     "}\n");
}

sub emit_writer {
    class_set_visibility('public');
    class_add ("    void store(handle_t _number, IntObject _value);\n");
    if (keys %g_writer_codes) {
        class_set_visibility('private');
        class_add ("    bool storeInternal(handle_t _number, IntObject _value);\n");
        push @c_output, ("\nbool\n",
                         "${g_class_name}::storeInternal(handle_t _number, IntObject _value)\n",
                         "{\n",
                         "    switch(_number) {");
        foreach (sort keys %g_writer_codes) {
            foreach (@{$g_writer_codes{$_}}) {
                push @c_output, "\n     case $_:";
            }
            push @c_output, " {\n        ", $_, "\n        return true;\n    }\n";
        }
        push @c_output, ("    }\n",
                         "    return false;\n",
                         "}\n");
        push @c_output, ("\nvoid\n",
                         "${g_class_name}::store(handle_t n, IntObject v)\n",
                         "{\n",
                         "    if (!storeInternal(n, v))\n",
                         "        throw IntException(\"Invalid assignment to '\" + string_t(property_names[n]) + \"'\");\n",
                         "}\n");
    } else {
        push @c_output, ("\nvoid\n",
                         "${g_class_name}::store(handle_t n, IntObject /*v*/)\n",
                         "{\n",
                         "    throw IntException(\"Invalid assignment to '\" + string_t(property_names[n]) + \"'\");\n",
                         "}\n");
    }
}

sub emit_default_functions {
    if (!exists $g_seen_functions{'clone'}) {
        class_set_visibility('public');
        class_add ("    ${g_class_name}* clone();\n");
        push @c_output, "\n// Create a copy of this object
${g_class_name}*
${g_class_name}::clone()
\{
    return new $g_class_name(*this);
\}\n";
    }

    if (!exists $g_seen_functions{'next'}) {
        class_set_visibility('public');
        class_add ("    bool next();\n");
        push @c_output, "\n// Next in list. Default implementation: no next object.
bool
${g_class_name}::next()
\{
    return false;
\}\n";
    }
}

sub emit_subs {
    class_set_visibility('public');
    class_add ("    void call(handle_t _number, int _nargs, IntExecutionContext& _xc);\n");

    push @c_output, "\n// Execute a subroutine / command call.\nvoid
${g_class_name}::call(handle_t _number, int _nargs, IntExecutionContext& _xc)
\{
    switch (_number) {\n";
    foreach (sort keys %g_sub_data) {
        push @c_output, "     case $g_sub_data{$_}{number}: {\n        // $_\n        ";
        push @c_output, $g_sub_data{$_}{code}, "\n        }\n        break;\n";
    }
    push @c_output, "     default:
        FAIL();
    }
}\n";
}

############################# Code Generation #############################

sub generate_codes {
    # Property accessors
    foreach (sort keys %g_name_data) {
        my $typ = $g_name_data{$_}->{type};
        my $reader = expand_code ($g_name_data{$_}->{reader_code}, $typ, 2,
                                  NAME => 'property_names[_number]');
        add_reader ($reader, $g_name_data{$_}->{number});
        if (defined $g_name_data{$_}->{writer_code}) {
            my $prefix = '';
            if (exists $g_type_accessor_code{$typ}) {
                $prefix = "$g_type_cxx_equivalent{$typ} _value_cooked = $g_type_accessor_code{$typ};\n";
            }
            my $writer = expand_code ($prefix . $g_name_data{$_}->{writer_code},
                                      $typ, 2,
                                      NAME => 'property_names[_number]',
                                      FAIL => 'return false',
                                      VALUE => '_value_cooked',
                                      VALUE_RAW => '_value');
            add_writer ($writer, $g_name_data{$_}->{number});
        }
    }

    # Subroutines
    foreach (sort keys %g_sub_data) {
        generate_code_for_sub($_, $g_sub_data{$_});
    }
}

sub generate_code_for_sub {
    my ($_name, $sub) = @_;
    my $indent = "        ";
    my $output_code = "\tif (!IF_argCheck(_nargs, ".$sub->{num_mandatory_args}.", ".
        scalar(@{$sub->{arg_names}}).", sub_names[_number], _xc))\n\t    return;\n";
    # Generate code for all parameters. We do it backwards, because that's
    # how we get the stack
    my @translations;
    for (my $i = scalar @{$sub->{arg_names}} - 1; $i >= 0; --$i) {
        my $type = $sub->{arg_types}[$i];
        my $name = $sub->{arg_names}[$i];
        my $default = $sub->{arg_defaults}[$i];
        push @translations, $name, $name;
        if (defined($type)) {
            # it has a defined type. If it's mandatory, we can be sure
            # it's there. Otherwise, check.
            if ($i >= $sub->{num_mandatory_args}) {
                die "no default for arg $i in $_name" if !defined $default; # parser makes sure this holds
                $output_code .= ("\t$g_type_cxx_equivalent{$type} $name ="
                                 ." _nargs > $i ? "
                                 .generate_type_accessor ($type, "_xc.popV()")
                                 ." : $default;\n");
            } else {
                $output_code .= "\t$g_type_cxx_equivalent{$type} $name = " . generate_type_accessor($type, "_xc.popV()") . ";\n";
            }
        } else {
            # variant arg.
            if ($i >= $sub->{num_mandatory_args}) {
                $output_code .= "\tIntObject $name = _nargs > $i ? _xc.popV() : IntObject();\n";
            } else {
                $output_code .= "\tIntObject $name = _xc.popV();\n";
            }
        }
    }
    $output_code =~ s/\t/$indent/g;
    $output_code .= $indent;
    $sub->{code} = expand_code ($output_code . $sub->{body},
                                '', 2,
                                NAME => 'sub_names[_number]',
                                @translations);
}

sub generate_type_accessor {
    my ($type, $repl) = @_;
    my $code = $g_type_accessor_code{$type} or die;
    $code =~ s/\b_value\b/$repl/g;
    $code;
}

sub add_reader {
    my ($code, $index) = @_;
    push @{$g_reader_codes{$code}}, $index;
}

sub add_writer {
    my ($code, $index) = @_;
    push @{$g_writer_codes{$code}}, $index;
}

############################ Expand a Function ############################

sub expand_add_translation {
    my ($name, $expansion) = @_;
    if (exists $g_unqual_translation{$name} && $g_unqual_translation{$name} ne $expansion) {
        $g_unqual_translation{$name} = '**AMBIGUOUS**';
    } else {
        $g_unqual_translation{$name} = $expansion;
    }
}

sub expand_code {
    # FIXME? Auto indent doesn't work yet. Doesn't really matter.
    my ($tx, $typ, $indent, @rest) = @_;
    my $output = '';
    lex_string (\$tx, 0);
    while (defined $lex_token) {
        if ($lex_token eq 'return') {
            lex_string (\$tx, 0);
            if (exists $g_type_cxx_equivalent{$typ}) {
                $output .= "return $g_type_cxx_equivalent{$typ}(" . expand_expr(\$tx, @rest) . ")";
            } else {
                $output .= "return ";
            }
        } elsif ($lex_token eq '{') {
            $indent++;
            $output .= '{';
            lex_string (\$tx, 0);
#            if ($lex_token ne '}') {
#                $output .= "{\n" . (' ' x (4*$indent));
#            } else {
#                $output .= "{";
#            }
        } elsif ($lex_token eq '}') {
            $indent--;
            $output .= '}';
#            $output =~ s/\s+\Z//s;
#            $output .= "\n" . (' ' x (4*$indent)) . '}';
            lex_string (\$tx, 1);
        } elsif ($lex_token eq ';') {
            $output .= $lex_token;
            lex_string (\$tx, 1);
        } else {
            $output .= expand_expr (\$tx, @rest);
        }
    }
    $output;
}

# Rudimentary expression parser: an expression is a list of tokens
# with balanced parens, up to a '{' or ';' or '}'.
sub expand_expr {
    my $ref = shift;
    my %alt_repl = @_;
    my $paren = 0;
    my @output;
    my $seen_qual = 0;
    while (defined($lex_token) && $lex_token ne '{' && $lex_token ne '}' && ($paren || $lex_token ne ';')) {
        # make sure that caller sees 'return'
        last if !$paren && $lex_token eq 'return';
        # is it an identifier which we want to expand?
        if (!$seen_qual && $lex_token =~ /^\w/) {
            if (exists $alt_repl{$lex_token}) {
                push @output, $alt_repl{$lex_token};
            } elsif (exists $g_unqual_translation{$lex_token}) {
                if ($g_unqual_translation{$lex_token} eq '**AMBIGUOUS**') {
                    die "'$lex_token' is ambiguous";
                }
                push @output, $g_unqual_translation{$lex_token};
            } else {
                push @output, $lex_token;
            }
            $seen_qual = 0;
        } else {
            push @output, $lex_token;
            if ($lex_token eq '::' || $lex_token eq '->' || $lex_token eq '.') {
                $seen_qual = 1;
            } elsif ($lex_token eq '(') {
                ++$paren;
                $seen_qual = 0;
            } elsif ($lex_token eq ')') {
                --$paren;
                $seen_qual = 0;
            } elsif ($lex_token =~ /\S/) {
                $seen_qual = 0;
            }
        }
        lex_string ($ref, 1);
    }
    join '', @output;
}

############################ Output Generation ############################

sub do_output {
    my $fn = shift;
    my $text = join "", @_;
# This was a hack for slow computers :-)))
#    if (open (CMP, "< $fn")) {
#        my $otext = join "", <CMP>;
#        close CMP;
#        if ($otext eq $text) {
#            print STDERR "$fn: unchanged\n";
#            return;
#        }
#    }
    open OUT, "> $fn" or die "$fn: $!";
    print OUT $text;
    close OUT;
}

################################## Lexer ##################################

sub lex_string {
    my ($line_ref, $report_ws) = @_;
    while (1) {
        if ($$line_ref =~ m{\G(\s+)}gc || $$line_ref =~ m{\G(//.*)}gc || $$line_ref =~ m{\G(/\*.*?\*/)}gc) {
            return $lex_token = $1
                if $report_ws;
        } elsif ($$line_ref =~ m{\G([a-zA-Z_\$][a-zA-Z0-9_\$.]*)}gc) {
            return $lex_token = $1;
        } elsif ($$line_ref =~ m{\G(\d+)}gc) {
            return $lex_token = $1;
        } elsif ($$line_ref =~ m{\G("([^\"\\]|\\.)*")}gc) {
            return $lex_token = $1;
        } elsif ($$line_ref =~ m{\G('([^\'\\]|\\.)*')}gc) {
            return $lex_token = $1;
        } elsif ($$line_ref =~ m{\G(%%|::|->)}gc) {
            # Note: all double-character tokens which we need must be
            # recognized here.
            return $lex_token = $1;
        } elsif ($$line_ref =~ m{\G(.)}gc) {
            return $lex_token = $1;
        } else {
            return $lex_token = undef;
        }
    }
}

sub lex_eof_nd {
    my $report_ws = shift;
    while (1) {
        if (!defined $lex_line || $lex_line =~ m|\G\s*$|gc) {
            $lex_line = <INPUT>;
            if (!defined $lex_line) {
                return $lex_token = '**EOF**';
            }
            chomp $lex_line;
            return $lex_token = "\n" if $report_ws;
        }
        defined (lex_string (\$lex_line, $report_ws))
            and return $lex_token;
    }
}

sub lex_eof {
    my $rv = lex_eof_nd(@_);
    # print STDERR $rv;
    $rv;
}

sub lex {
    lex_eof (@_);
    die "unexpected end-of-file"
        if $lex_token eq '**EOF**';
    die "unexpected section delimiter"
        if $lex_token eq '%%';
    $lex_token;
}

# Assuming current token is a '{', read a function body. When starting
# at a ':', read that, too.
sub lex_function_body {
    my $indent = shift || '';
    my $total = $lex_token;
    my $brace_level = 1;
    if ($lex_token eq ':') {
        do { $total .= lex(1); } while ($lex_token ne '{');
    }
    while ($brace_level) {
        lex (1);
        if ($lex_token eq '{') { ++$brace_level; }
        if ($lex_token eq '}') { --$brace_level; }
        $total .= $lex_token;
        $total .= $indent if $lex_token eq '\n';
    }
    $total;
}

# Assuming current token is first of an expression, read the expression.
# Leaves pointer at terminating ',', ';' or ')'.
sub lex_expr {
    my $output = "";
    my $paren_level = 0;
    while (1) {
        if ($lex_token eq '(') {
            ++$paren_level;
        } elsif ($lex_token eq ')') {
            return $output
                if $paren_level == 0;
            --$paren_level;
        } elsif ($lex_token eq ';' || $lex_token eq ',') {
            return $output
                if $paren_level == 0;
        }
        $output .= $lex_token;
        lex(1);
    }
}

sub lex_expect {
    foreach (@_) {
        if ($lex_token ne $_) { die "expected '$_', got '$lex_token'"; }
        lex (0);
    }
}
