/**
  *  \file ui/dlg-fileselect.cc
  *
  *  (c) 2001-2007 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "ui/dlg-fileselect.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "util/nls.h"
#include "ui/msgbox.h"


/*! \class UIFileList
    \brief File Selection List

    List widget displaying a file selection list. */

/** Constructor.
    \param id        Widget id
    \param directory Initial directory
    \param wildcard  Initial filter
    \param lines     Number of lines to display */
UIFileList::UIFileList(int id, Ptr<Directory> directory, string_t wildcard, int lines)
    : UIListbox(id, lines, 20, 0),
      wildcard(wildcard), directory(directory)
{
    update(directory);
}

UIFileList::~UIFileList()
{ }

static bool
sortDirectory(DirectoryEntry* a, DirectoryEntry* b)
{
    if (a == 0)
        return true;

    if (b == 0)
        return false;

    if (a->getFileType() != b->getFileType())
        return a->getFileType() == DirectoryEntry::FT_DIRECTORY;

    return a->getBaseName() < b->getBaseName();
}

/** Change to directory and update content.
    \param dir New directory
    \throws FileOpenException / std::runtime_error if directory cannot
    be read; in this case, the change does not take place. */
void
UIFileList::update(Ptr<Directory> dir)
{
    pat.setPattern(wildcard);

    std::vector<Ptr<DirectoryEntry> > new_content;

    if (Ptr<Directory> p = dir->getParent())
        new_content.push_back(0);

    Ptr<Enumeration<Ptr<DirectoryEntry> > > e = dir->getEntries();
    while (e->hasMoreElements()) {
        Ptr<DirectoryEntry> entry = e->getNextElement();
        if (entry->getFileType() == DirectoryEntry::FT_DIRECTORY)
            new_content.push_back(entry);
        else if (pat.match(entry->getBaseName()))
            new_content.push_back(entry);
    }

    std::sort(new_content.begin(), new_content.end(), sortDirectory);

    content.swap(new_content);
    directory = dir;

    setSize(content.size());
    setCurrent(0);
    adjustPosition();
    drawWidget();
    sig_moved.raise(getId());
}

void
UIFileList::drawPart(GfxCanvas& can, int from, int to)
{
    getSkin().startDraw();

    unsigned index = from + getTop();
    for (int i = from; i < to; ++i) {
        if (index < content.size()) {
            Ptr<DirectoryEntry> p = content[index];
            if (p == 0) {
                drawEntry(can, UIColor::tc_Static, i, _("[Up one level]"));
                ++index;
            } else {
                string_t line = " " + p->getBaseName();
                if (p->getFileType() == DirectoryEntry::FT_DIRECTORY)
                    line[0] = Directory::PATH_SEPARATOR;
                drawEntry(can, UIColor::tc_Static, i, line);
                ++index;
            }
        } else {
            drawEntry(can, UIColor::tc_Static, i, string_t());
        }
    }

}

/** Get string representing current selection. */
string_t
UIFileList::getCurrentSelection() const
{
    unsigned index = getCurrent();
    if (index < content.size()) {
        Ptr<DirectoryEntry> p = content[index];
        if (p == 0) {
            return Directory::makeFileName(directory->getParent()->getName(), wildcard);
        } else if (p->getFileType() == DirectoryEntry::FT_DIRECTORY) {
            if (directory->getName().size() != 0) {
                return Directory::makeFileName(p->getBaseName(), wildcard);
            } else {
                return Directory::makeFileName(p->getName(), wildcard);
            }
        } else {
            return p->getBaseName();
        }
    } else {
        return string_t();
    }
}

/** Handle input. */
bool
UIFileList::handleInput(string_t input)
{
    /* first, split into path and possible pattern */
    string_t basename = Directory::getFileName(input);
    string_t dirname = Directory::getDirName(input);

    string_t current_dir = directory->getName();
    if (current_dir.size() == 0) {
        /* we don't have a current directory */
        if (!Directory::isAbsoluteName(dirname)) {
            /* FIXME: message? */
            return false;
        } else {
            /* ok */
        }
    } else {
        /* we have a current directory */
        dirname = Directory::makeFileName(current_dir, dirname);
    }

    pat.setPattern(basename);
    if (!pat.hasWildcard()) {
        /* we have a result */
        /* FIXME: in case user gave us a directory name, we should
           switch there instead of returning it as-is */
        result = Directory::makeFileName(dirname, pat.getFileName());
        return true;
    } else {
        /* change directory */
        /* FIXME: check that it exists */
        Ptr<Directory> dir = Directory::openDir(dirname);
        try {
            wildcard = basename;
            update(dir);
        }
        catch (std::runtime_error& e) {
            messageBox(e.what(), _("Error"));
        }
        return false;
    }
}

/**************************** UIFileListWindow ***************************/

UIFileListWindow::UIFileListWindow(string_t title, string_t wildcard)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      list(1000, Directory::openDir(Directory::getWorkingDirectory()), wildcard, 20),
      input(1000, 20, ss_Alt + 'n', 0, 1001)
{
    UIGroup& group_input   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group_list    = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    group_input.add(h.add(new UIStaticText(_("Name:"), UIStaticText::BlackTitle)));
    group_input.add(input);

    group_list.add(list);
    group_list.add(h.add(new UIScrollbar(list)));

    UIButton& button_ok     = h.add(new UIButton(1, _("OK"), SDLK_RETURN, UIButton::aNothing));
    UIButton& button_cancel = h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE, UIButton::aNothing));

    group_buttons.add(button_ok);
    group_buttons.add(button_cancel);
    group_buttons.add(h.add(new UISpacer()));

    setLayout(UIVBoxLayout::instance5);
    add(group_input);
    add(group_list);
    add(group_buttons);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab)));

    pack();

    button_ok.sig_fired.add(this, &UIFileListWindow::onOK);
    button_cancel.sig_fired.add(this, &UIFileListWindow::onCancel);
    list.sig_moved.add(this, &UIFileListWindow::onMove);
}

UIFileListWindow::~UIFileListWindow()
{ }

string_t
UIFileListWindow::getSelectedFile() const
{
    return list.getResult();
}

void
UIFileListWindow::onOK()
{
    if (list.handleInput(input.getText()))
        stop(1);
}

void
UIFileListWindow::onCancel()
{
    stop(0);
}

void
UIFileListWindow::onMove()
{
    input.setText(list.getCurrentSelection());
}
